@@ -1,7 +1,7 @@
 /*
  * lftp - file transfer program
  *
- * Copyright (c) 1996-2012 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1996-2016 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -51,6 +51,7 @@ public:
    xstring  name;
    xstring  longname;
    xstring_c symlink;
+   xstring_c uri;
    mode_t   mode;
    FileTimestamp date;
    off_t    size;
@@ -63,7 +64,8 @@ public:
       UNKNOWN=0,
       DIRECTORY,
       SYMLINK,
-      NORMAL
+      NORMAL,
+      REDIRECT,
    };
    type	 filetype;
 
@@ -85,7 +87,8 @@ public:
    void Init();
    FileInfo() { Init(); }
    FileInfo(const FileInfo &fi);
-   FileInfo(const char *n);
+   FileInfo(const char *n) { Init(); SetName(n); }
+   FileInfo(const xstring& n) { Init(); SetName(n); }
    ~FileInfo();
 
    void SetName(const char *n) { name.set(n); def(NAME); }
@@ -100,10 +103,13 @@ public:
    void SetDate(time_t t,int prec) { date.set(t,prec); def(DATE); }
    void SetType(type t) { filetype=t; def(TYPE); }
    void SetSymlink(const char *s) { symlink.set(s); filetype=SYMLINK; def(TYPE|SYMLINK_DEF); }
+   void SetRedirect(const char *s) { symlink.set(s); filetype=REDIRECT; def(TYPE|SYMLINK_DEF); }
+   const char *GetRedirect() const { return symlink; }
    void	SetSize(off_t s) { size=s; def(SIZE); }
    void	SetNlink(int n) { nlinks=n; def(NLINKS); }
 
    void	 Merge(const FileInfo&);
+   void  MergeInfo(const FileInfo& f,unsigned mask);
 
    bool	 SameAs(const FileInfo *,int ignore) const;
    bool	 OlderThan(time_t t) const;
@@ -137,19 +143,22 @@ class PatternSet;
 class FileSet
 {
 public:
-   enum sort_e { BYNAME, BYSIZE, DIRSFIRST, BYRANK, BYDATE };
+   enum sort_e { BYNAME, BYSIZE, DIRSFIRST, BYRANK, BYDATE, BYNAME_FLAT };
 
 private:
    RefArray<FileInfo> files;
 
    /* indexes when sort != NAME: */
    xarray<int> sorted;
+   sort_e sort_mode;
 
    int	 ind;
 
    void	 Sub(int);
+   FileInfo *Borrow(int);
 
    void add_before(int pos,FileInfo *fi);
+   void assert_sorted() const;
 
 public:
    FileSet();
@@ -165,6 +174,7 @@ public:
 
    void	 Add(FileInfo *);
    void	 Merge(const FileSet *);
+   void	 Merge_insert(const FileSet *set);
    void	 SubtractSame(const FileSet *,int ignore);
    void	 SubtractAny(const FileSet *);
    void  SubtractTimeCmp(bool (FileInfo::*cmp)(time_t) const,time_t);
@@ -180,24 +190,27 @@ public:
    void  SubtractDirs(const FileSet *);
    void  SubtractNotOlderDirs(const FileSet *);
    void  SubtractCurr();
+   bool  SubtractByName(const char *name);
    void  Sort(sort_e newsort, bool casefold=false, bool reverse=false);
    void  Unsort();
    void	 SortByPatternList(const char *list_c);
    void	 ReverseSort();
+   void	 UnsortFlat();
 
-   void	 Exclude(const char *prefix,const PatternSet *x);
+   void	 Exclude(const char *prefix,const PatternSet *x,FileSet *fsx=0);
    void	 ExcludeDots();
    void	 ExcludeCompound();
-   void	 ExcludeUnaccessible();
+   void	 ExcludeUnaccessible(const char *user=0);
 
    void	 rewind() { ind=0; }
    FileInfo *curr();
    FileInfo *next();
+   FileInfo *borrow_curr() { return Borrow(ind--); }
 
    void	 LocalRemove(const char *dir);
-   void	 LocalUtime(const char *dir,bool only_dirs=false);
-   void	 LocalChmod(const char *dir,mode_t mask=0);
-   void	 LocalChown(const char *dir);
+   void	 LocalUtime(const char *dir,bool only_dirs=false,bool flat=false);
+   void	 LocalChmod(const char *dir,mode_t mask=0,bool flat=false);
+   void	 LocalChown(const char *dir,bool flat=false);
 
    void Count(int *d,int *f,int *s,int *o) const;
    void CountBytes(long long *b) const;
